import functools
import warnings
from typing import Callable, Optional

try:
    from numba import njit, typed, types

except ImportError:

    warnings.warn(
        "numba not installed, falling back to stubs. "
        "Install numba for better napari-graph performance."
    )

    def njit(func: Optional[Callable] = None, **kwargs) -> Callable:
        """Immitate numba.njit decorator"""

        def _decorator(f: Callable) -> Callable:
            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                return f(*args, **kwargs)

            return wrapper

        if func:
            return _decorator(func)

        return _decorator

    class StubList(list):
        @staticmethod
        def empty_list(type) -> list:
            return []

    class StubDict(dict):
        @staticmethod
        def empty(key_type, value_type) -> dict:
            return {}

    class typed:  # type: ignore[no-redef]
        List = StubList
        Dict = StubDict

    class types:  # type: ignore[no-redef]
        int64 = int



# napari-graph documentation build configuration file
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# Add the local code to the Python path, so docs are generated for
# current working copy

from typing import Any, Dict

import napari_graph

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Execution options
execution_timeout = 300  # in seconds?

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx.ext.napoleon',
    'sphinxcontrib.bibtex',
    'sphinx_copybutton',
    'myst_nb',
]

# sphinx-copybutton prompt-stripping configuration
copybutton_prompt_text = (
    r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
)
copybutton_prompt_is_regexp = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'napari-graph'
copyright = '2022, Jordão Bragantini & Juan Nunez-Iglesias'
author = 'Jordão Bragantini & Juan Nunez-Iglesias'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.

version = napari_graph.__version__
# The full version, including alpha/beta/rc tags.
release = napari_graph.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# List of publications
bibtex_bibfiles = ['publications.bib']

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = 'sphinx_material'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#

html_theme = 'sphinx_material'

# Material theme options (see theme.conf for more information)
html_theme_options = {
    # Set the name of the project to appear in the navigation.
    'nav_title': 'napari-graph',
    # Set you GA account ID to enable tracking
    # 'google_analytics_account': 'UA-XXXXX',
    # Specify a base_url used to generate sitemap.xml. If not
    # specified, then no sitemap will be built.
    'base_url': 'https://napari.org/napari-graph',
    # Set the color and the accent color
    'color_primary': 'blue',
    'color_accent': 'light-blue',
    # Set the repo location to get a badge with stats
    'repo_url': 'https://github.com/napari/napari-graph',
    'repo_name': 'Skan',
    # Visible levels of the global TOC; -1 means unlimited
    'globaltoc_depth': 1,
    # If False, expand all TOC entries
    'globaltoc_collapse': False,
    # If True, show hidden TOC entries
    'globaltoc_includehidden': False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    "**": [
        "logo-text.html",
        "globaltoc.html",
        "localtoc.html",
        "searchbox.html",
    ]
}

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'napari-graph-docs'

# -- Options for LaTeX output ---------------------------------------------

latex_elements: Dict[str, Any] = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        'napari-graph.tex',
        'napari-graph documentation',
        author,
        'manual',
    ),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'napari-graph-documentation', [author], 1)]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        'napari-graph',
        'napari-graph documentation',
        author,
        'napari-graph',
        'Fast editable graphs in Python',
        'Miscellaneous',
    ),
]



from typing import List, Optional, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike

from napari_graph.base_graph import (
    _EDGE_EMPTY_PTR,
    _NODE_EMPTY_PTR,
    BaseGraph,
    _iterate_edges,
    _remove_edge,
)
from napari_graph.numba import njit, typed
from napari_graph.undirected_graph import _UN_EDGE_SIZE

"""
Directed edge constants for accessing the directed graph buffer data.
Each edge occupies _DI_EDGE_SIZE spaces on the graph buffer.
_LL_DI_EDGE_POS indicates the displacement between the edge initial index and
the **target** edge directed linked list position.

Example of a directed graph edge buffer:
[
    source_node_buffer_id_0,
    target_node_buffer_id_0,
    source_edge_linked_list_0,
    target_edge_linked_List_0,
    source_node_buffer_id_1,
    target_node_buffer_id_1,
    source_edge_linked_list_1,
    target_edge_linked_List_1,
    ...
]
"""
_DI_EDGE_SIZE = _UN_EDGE_SIZE + 1
_LL_DI_EDGE_POS = 2


@njit
def _add_directed_edge(
    buffer: np.ndarray,
    node2src_edges: np.ndarray,
    node2tgt_edges: np.ndarray,
    empty_idx: int,
    src_node: int,
    tgt_node: int,
) -> int:
    """Add a single directed edge to `buffer`.

    This updates the `buffer`'s source and target linked list
    and the nodes to edges mappings.

    NOTE: see `_add_undirected_edge` docs for comment about cache misses.

    Parameters
    ----------
    buffer : np.ndarray
        Edges buffer.
    node2src_edges : np.ndarray
        Mapping from node indices to source edge buffer indices -- head of edges linked list.
    node2tgt_edges : np.ndarray
        Mapping from node indices to target edge buffer indices -- head of edges linked list.
    empty_idx : int
        First index of empty edges linked list.
    src_node : int
        Source node of added edge.
    tgt_node : int
        Target node of added edge.

    Returns
    -------
    int
        New first index of empty edges linked list.
    """
    if empty_idx == _EDGE_EMPTY_PTR:
        raise ValueError("Edge buffer is full.")

    elif empty_idx < 0:
        raise ValueError("Invalid empty index.")

    next_src_edge = node2src_edges[src_node]
    next_tgt_edge = node2tgt_edges[tgt_node]
    node2src_edges[src_node] = empty_idx
    node2tgt_edges[tgt_node] = empty_idx

    buffer_index = empty_idx * _DI_EDGE_SIZE
    next_empty = buffer[buffer_index + _LL_DI_EDGE_POS]

    buffer[buffer_index] = src_node
    buffer[buffer_index + 1] = tgt_node
    buffer[buffer_index + _LL_DI_EDGE_POS] = next_src_edge
    buffer[buffer_index + _LL_DI_EDGE_POS + 1] = next_tgt_edge

    return next_empty


@njit
def _add_directed_edges(
    buffer: np.ndarray,
    edges: np.ndarray,
    empty_idx: int,
    n_edges: int,
    node2src_edges: np.ndarray,
    node2tgt_edges: np.ndarray,
) -> Tuple[int, int]:
    """Add an array of edges into the `buffer`.

    Directed edges contains two linked lists, outgoing (source) and incoming
    (target) edges.
    """
    size = edges.shape[0]
    for i in range(size):

        empty_idx = _add_directed_edge(
            buffer,
            node2src_edges,
            node2tgt_edges,
            empty_idx,
            edges[i, 0],
            edges[i, 1],
        )
        n_edges += 1

    return empty_idx, n_edges


@njit(inline='always')
def _remove_target_edge(
    src_node: int,
    tgt_node: int,
    edges_buffer: np.ndarray,
    node2tgt_edges: np.ndarray,
) -> None:
    """Remove edge from target edges linked list.
    It doesn't clean the buffer, because it'll be used later.

    Parameters
    ----------
    src_node : int
        Source node of added edge.
    tgt_node : int
        Target node of added edge.
    edges_buffer : np.ndarray
        Edges buffer.
    node2tgt_edges : np.ndarray
        Mapping from node indices to target edge buffer indices -- head of edges linked list.
    """
    idx = node2tgt_edges[tgt_node]  # different indexing from source edge
    prev_buffer_idx = _EDGE_EMPTY_PTR

    for _ in range(edges_buffer.shape[0] // _DI_EDGE_SIZE):
        if idx == _EDGE_EMPTY_PTR:
            raise ValueError(
                "Could not find target node at directed edge removal."
            )

        buffer_idx = idx * _DI_EDGE_SIZE
        next_edge_idx = edges_buffer[buffer_idx + _LL_DI_EDGE_POS + 1]

        # edge found
        if (
            edges_buffer[buffer_idx] == src_node
        ):  # different indexing from source edge
            # skipping found edge from linked list
            if prev_buffer_idx == _EDGE_EMPTY_PTR:
                node2tgt_edges[
                    tgt_node
                ] = next_edge_idx  # different indexing from source edge
            else:
                edges_buffer[
                    prev_buffer_idx + _LL_DI_EDGE_POS + 1
                ] = next_edge_idx

            edges_buffer[buffer_idx + _LL_DI_EDGE_POS + 1] = _EDGE_EMPTY_PTR
            break

        # moving to next edge
        idx = next_edge_idx
        prev_buffer_idx = buffer_idx
    else:
        raise ValueError(
            "Infinite loop detected at target edge removal, edges buffer must be corrupted."
        )


@njit
def _remove_directed_edge(
    src_node: int,
    tgt_node: int,
    empty_idx: int,
    edges_buffer: np.ndarray,
    node2src_edges: np.ndarray,
    node2tgt_edges: np.ndarray,
) -> int:
    """Remove a single directed edge from the edges buffer."""

    # must be executed before default edge removal and cleanup
    _remove_target_edge(src_node, tgt_node, edges_buffer, node2tgt_edges)

    empty_idx = _remove_edge(
        src_node,
        tgt_node,
        empty_idx,
        edges_buffer,
        node2src_edges,
        _DI_EDGE_SIZE,
        _LL_DI_EDGE_POS,
    )

    return empty_idx


@njit
def _remove_directed_edges(
    edges: np.ndarray,
    empty_idx: int,
    n_edges: int,
    edges_buffer: np.ndarray,
    node2src_edges: np.ndarray,
    node2tgt_edges: np.ndarray,
) -> Tuple[int, int]:
    """Remove an array of edges from the edges buffer."""

    for i in range(edges.shape[0]):
        empty_idx = _remove_directed_edge(
            edges[i, 0],
            edges[i, 1],
            empty_idx,
            edges_buffer,
            node2src_edges,
            node2tgt_edges,
        )
        n_edges -= 1
    return empty_idx, n_edges


@njit
def _remove_directed_incident_edges(
    node: int,
    empty_idx: int,
    n_edges: int,
    edges_buffer: np.ndarray,
    node2src_edges: np.ndarray,
    node2tgt_edges: np.ndarray,
    is_target: int,
) -> Tuple[int, int]:
    """Remove directed edges from the buffer that contain the given `node`.

    The parameter `is_target` should be zero to remove edges where `node` is
    the source node, and 1 for the target node.

    Parameters
    ----------
    node : int
        Node index in the buffer domain.
    empty_idx : int
        Index first edge (head of) linked list
    n_edges : int
        Current number of total edges
    edges_buffer : np.ndarray
        Buffer containing the edges data
    node2src_edges : np.ndarray
        Mapping from node indices to source edge buffer indices -- head of edges linked list.
    node2tgt_edges : np.ndarray
        Mapping from node indices to target edge buffer indices -- head of edges linked list.
    is_target : int
        Binary integer flag indicating if it is a target or not, used to shift linked list position.

    Returns
    -------
    Tuple[int, int]
        New empty linked list head, new number of edges
    """
    if is_target:
        idx = node2tgt_edges[node]
    else:
        idx = node2src_edges[node]

    # safe guard against a corrupted buffer causing an infite loop
    for _ in range(edges_buffer.shape[0] // _DI_EDGE_SIZE):
        if idx == _EDGE_EMPTY_PTR:
            break  # no edges left at the given node

        buffer_idx = idx * _DI_EDGE_SIZE
        next_idx = edges_buffer[buffer_idx + _LL_DI_EDGE_POS + is_target]

        src_node = edges_buffer[buffer_idx]
        tgt_node = edges_buffer[buffer_idx + 1]

        # must be removed before source edge, due to information cleanup
        _remove_target_edge(src_node, tgt_node, edges_buffer, node2tgt_edges)
        empty_idx = _remove_edge(
            src_node,
            tgt_node,
            empty_idx,
            edges_buffer,
            node2src_edges,
            _DI_EDGE_SIZE,
            _LL_DI_EDGE_POS,
        )

        idx = next_idx
        n_edges -= 1
    else:
        if idx != _EDGE_EMPTY_PTR:
            raise ValueError(
                "Infinite loop detected at directed graph node removal, edges buffer must be corrupted."
            )

    return empty_idx, n_edges


@njit
def _iterate_directed_source_edges(
    edge_ptr_indices: np.ndarray, edges_buffer: np.ndarray
) -> typed.List:
    """Inline the edges size and linked list position shift."""
    return _iterate_edges(
        edge_ptr_indices, edges_buffer, _DI_EDGE_SIZE, _LL_DI_EDGE_POS
    )


@njit
def _iterate_directed_target_edges(
    edge_ptr_indices: np.ndarray, edges_buffer: np.ndarray
) -> typed.List:
    """Inline the edges size and linked list position shift."""
    return _iterate_edges(
        edge_ptr_indices, edges_buffer, _DI_EDGE_SIZE, _LL_DI_EDGE_POS + 1
    )


class DirectedGraph(BaseGraph):
    """Directed graph class.

    Parameters
    ----------
    edges : ArrayLike
        Nx2 array of pair of nodes (edges).
    coords :
        Optional array of spatial coordinates of nodes.
    ndim : int
        Number of spatial dimensions of graph.
    n_nodes : int
        Optional number of nodes to pre-allocate in the graph.
    n_edges : int
        Optional number of edges to pre-allocate in the graph.
    """

    _EDGE_DUPLICATION = 1
    _EDGE_SIZE = _DI_EDGE_SIZE
    _LL_EDGE_POS = _LL_DI_EDGE_POS

    def _init_node_buffers(self, n_nodes: int) -> None:
        super()._init_node_buffers(n_nodes)
        self._node2tgt_edges = np.full(
            n_nodes, fill_value=_EDGE_EMPTY_PTR, dtype=int
        )

    def _realloc_nodes_buffers(self, size: int) -> None:
        diff_size = size - self.n_allocated_nodes
        super()._realloc_nodes_buffers(size)
        self._node2tgt_edges = np.append(
            self._node2tgt_edges,
            np.full(diff_size, fill_value=_NODE_EMPTY_PTR, dtype=np.int64),
        )

    def _add_edges(self, edges: np.ndarray) -> None:
        self._empty_edge_idx, self._n_edges = _add_directed_edges(
            self._edges_buffer,
            edges,
            self._empty_edge_idx,
            self._n_edges,
            self._node2edges,
            self._node2tgt_edges,
        )

    def get_edges(
        self, nodes: Optional[ArrayLike] = None, mode: str = 'indices'
    ) -> Union[List[np.ndarray], np.ndarray]:
        """`source_edges` alias"""
        return self.get_target_edges(nodes, mode)

    def out_edges(
        self, nodes: Optional[ArrayLike] = None, mode: str = 'indices'
    ) -> Union[List[np.ndarray], np.ndarray]:
        """`source_edges` alias"""
        return self.get_source_edges(nodes, mode)

    def get_source_edges(
        self, nodes: Optional[ArrayLike] = None, mode: str = 'indices'
    ) -> Union[List[np.ndarray], np.ndarray]:
        """Return the source edges (outgoing) of the given nodes.

        If no nodes are provided, all source edges are returned.

        Parameters
        ----------
        nodes : Optional[ArrayLike], optional
            Node indices, by default None
        mode : str
            Type of data queried from the edges. For example, `indices` or
            `coords`.

        Returns
        -------
        List[np.ndarray]
            List of (N_i) x 2 x D arrays, where N_i is the number of edges at
            the ith node.  D is the dimensionality of `coords` when
            mode == `coords` and is ignored when mode == `indices`. N_i
            dimension is ignored when N_i is 1.
        """
        return self._iterate_edges_generic(
            nodes,
            node2edges=self._node2edges,
            iterate_edges_func=_iterate_directed_source_edges,
            mode=mode,
        )

    def in_edges(
        self, nodes: Optional[ArrayLike] = None, mode: str = 'indices'
    ) -> Union[List[np.ndarray], np.ndarray]:
        """`target_edges` alias"""
        return self.get_target_edges(nodes, mode)

    def get_target_edges(
        self, nodes: Optional[ArrayLike] = None, mode: str = 'indices'
    ) -> Union[List[np.ndarray], np.ndarray]:
        """Return the target edges (incoming) of the given nodes.

        If no nodes are provided, all target edges are returned.

        Parameters
        ----------
        nodes : Optional[ArrayLike], optional
            Node indices, by default None
        mode : str
            Type of data queried from the edges. For example, `indices` or
            `coords`.

        Returns
        -------
        List[np.ndarray]
            List of (N_i) x 2 x D arrays, where N_i is the number of edges at
            the ith node.  D is the dimensionality of `coords` when
            mode == `coords` and it's ignored when mode == `indices`. N_i
            dimension is ignored when N_i is 1.
        """
        return self._iterate_edges_generic(
            nodes,
            node2edges=self._node2tgt_edges,
            iterate_edges_func=_iterate_directed_target_edges,
            mode=mode,
        )

    def _remove_edges(self, edges: np.ndarray) -> None:
        self._empty_edge_idx, self._n_edges = _remove_directed_edges(
            edges,
            self._empty_edge_idx,
            self._n_edges,
            self._edges_buffer,
            self._node2edges,
            self._node2tgt_edges,
        )

    def _remove_incident_edges(self, node_buffer_index: int) -> None:
        """Remove directed edges that contain `node` in either direction."""
        for is_target in (1, 0):
            (
                self._empty_edge_idx,
                self._n_edges,
            ) = _remove_directed_incident_edges(
                node_buffer_index,
                self._empty_edge_idx,
                self._n_edges,
                self._edges_buffer,
                self._node2edges,  # node2src_edges
                self._node2tgt_edges,
                is_target,
            )



from typing import List, Optional, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike

from napari_graph.base_graph import (
    _EDGE_EMPTY_PTR,
    BaseGraph,
    _iterate_edges,
    _remove_edge,
)
from napari_graph.numba import njit, typed

"""
Undirected edge constants for accessing the directed graph buffer data.
Each edge occupies _UN_EDGE_SIZE spaces on the graph buffer.
_LL_UN_EDGE_POS indicates the displacement between the edge initial index and
the edge undirected linked list position.

Example of a directed graph edge buffer:
[
    source_node_buffer_id_0,
    target_node_buffer_id_0,
    edge_linked_list_0,
    source_node_buffer_id_1,
    target_node_buffer_id_1,
    edge_linked_list_1,
    ...
]
"""
_UN_EDGE_SIZE = 3
_LL_UN_EDGE_POS = 2


@njit
def _add_undirected_edge(
    buffer: np.ndarray,
    node2edges: np.ndarray,
    empty_idx: int,
    src_node: int,
    tgt_node: int,
) -> int:
    """Add a single edge (`src_idx`, `tgt_idx`) to `buffer`.

    Update the edge linked list (present in the buffer) and the `node2edges`
    mapping (head of linked list).

    NOTE: Edges are added at the beginning of the linked list so we don't have
    to track its tail and the operation can be done in O(1). This might
    decrease cache hits because they're sorted in memory in the opposite
    direction we iterate it.

    Parameters
    ----------
    buffer : np.ndarray
        Edges buffer.
    node2edges : np.ndarray
        Mapping from node indices to edge buffer indices -- head of edges linked list.
    empty_idx : int
        First index of empty edges linked list.
    src_node : int
        Source node of added edge.
    tgt_node : int
        Target node of added edge.

    Returns
    -------
    int
        New first index of empty edges linked list.
    """

    if empty_idx == _EDGE_EMPTY_PTR:
        raise ValueError("Edge buffer is full.")

    elif empty_idx < 0:
        raise ValueError("Invalid empty index.")

    next_edge = node2edges[src_node]
    node2edges[src_node] = empty_idx

    buffer_index = empty_idx * _UN_EDGE_SIZE
    next_empty = buffer[buffer_index + _LL_UN_EDGE_POS]

    buffer[buffer_index] = src_node
    buffer[buffer_index + 1] = tgt_node
    buffer[buffer_index + _LL_UN_EDGE_POS] = next_edge

    return next_empty


@njit
def _add_undirected_edges(
    buffer: np.ndarray,
    edges: np.ndarray,
    empty_idx: int,
    n_edges: int,
    node2edges: np.ndarray,
) -> Tuple[int, int]:
    """Add an array of edges into the `buffer`.

    Edges are duplicated so both directions are available for fast graph
    transversal.
    """
    size = edges.shape[0]
    for i in range(size):

        # adding (u, v)
        empty_idx = _add_undirected_edge(
            buffer, node2edges, empty_idx, edges[i, 0], edges[i, 1]
        )
        # adding (v, u)
        empty_idx = _add_undirected_edge(
            buffer, node2edges, empty_idx, edges[i, 1], edges[i, 0]
        )

        n_edges += 1

    return empty_idx, n_edges


@njit
def _remove_undirected_edge(
    src_node: int,
    tgt_node: int,
    empty_idx: int,
    edges_buffer: np.ndarray,
    node2edges: np.ndarray,
) -> int:
    """Remove a single edge (and its duplicated sibling edge) from the buffer.

    NOTE: Edges are removed such that empty pairs are consecutive in memory.
    """
    empty_idx = _remove_edge(
        tgt_node,
        src_node,
        empty_idx,
        edges_buffer,
        node2edges,
        _UN_EDGE_SIZE,
        _LL_UN_EDGE_POS,
    )

    empty_idx = _remove_edge(
        src_node,
        tgt_node,
        empty_idx,
        edges_buffer,
        node2edges,
        _UN_EDGE_SIZE,
        _LL_UN_EDGE_POS,
    )

    return empty_idx


@njit
def _remove_undirected_edges(
    edges: np.ndarray,
    empty_idx: int,
    n_edges: int,
    edges_buffer: np.ndarray,
    node2edges: np.ndarray,
) -> Tuple[int, int]:
    """Remove an array of edges (and their duplicated siblings) from buffer."""

    for i in range(edges.shape[0]):
        empty_idx = _remove_undirected_edge(
            edges[i, 0], edges[i, 1], empty_idx, edges_buffer, node2edges
        )
        n_edges -= 1
    return empty_idx, n_edges


@njit
def _remove_undirected_incident_edges(
    node: int,
    empty_idx: int,
    n_edges: int,
    edges_buffer: np.ndarray,
    node2edges: np.ndarray,
) -> Tuple[int, int]:
    """Removes every edges that contains `node_idx`.

    NOTE: Edges are removed such that empty pairs are consecutive in memory.

    Parameters
    ----------
    node : int
        Node index in the buffer domain.
    empty_idx : int
        Index first edge (head of) linked list
    n_edges : int
        Current number of total edges
    edges_buffer : np.ndarray
        Buffer containing the edges data
    node2edges : np.ndarray
        Mapping from node indices to edge buffer indices -- head of edges linked list.

    Returns
    -------
    Tuple[int, int]
        New empty linked list head, new number of edges
    """

    # the edges are removed such that the empty edges linked list contains
    # two positions adjacent in memory so we can serialize the edges using
    # numpy vectorization
    idx = node2edges[node]

    for _ in range(edges_buffer.shape[0] // _UN_EDGE_SIZE):
        if idx == _EDGE_EMPTY_PTR:
            break  # no edges left at the given node

        buffer_idx = idx * _UN_EDGE_SIZE
        next_idx = edges_buffer[buffer_idx + _LL_UN_EDGE_POS]
        # checking if sibling edges is before or after current node
        if (
            buffer_idx > 0
            and edges_buffer[buffer_idx - _UN_EDGE_SIZE + 1]
            == edges_buffer[buffer_idx]
            and edges_buffer[buffer_idx - _UN_EDGE_SIZE]
            == edges_buffer[buffer_idx + 1]
        ):

            src_node = edges_buffer[buffer_idx + 1]
            tgt_node = edges_buffer[buffer_idx]
        else:
            src_node = edges_buffer[buffer_idx]
            tgt_node = edges_buffer[buffer_idx + 1]

        empty_idx = _remove_edge(
            tgt_node,
            src_node,
            empty_idx,
            edges_buffer,
            node2edges,
            _UN_EDGE_SIZE,
            _LL_UN_EDGE_POS,
        )

        empty_idx = _remove_edge(
            src_node,
            tgt_node,
            empty_idx,
            edges_buffer,
            node2edges,
            _UN_EDGE_SIZE,
            _LL_UN_EDGE_POS,
        )

        idx = next_idx
        n_edges -= 1
    else:
        if idx != _EDGE_EMPTY_PTR:
            raise ValueError(
                "Infinite loop detected at undirected graph node removal, edges buffer must be corrupted."
            )

    return empty_idx, n_edges


@njit
def _iterate_undirected_edges(
    edge_ptr_indices: np.ndarray, edges_buffer: np.ndarray
) -> typed.List:
    """Inline the edges size and linked list position shift."""
    return _iterate_edges(
        edge_ptr_indices, edges_buffer, _UN_EDGE_SIZE, _LL_UN_EDGE_POS
    )


class UndirectedGraph(BaseGraph):
    """Undirected graph class.

    Parameters
    ----------
    n_nodes : int
        Number of nodes to allocate to graph.
    ndim : int
        Number of spatial dimensions of graph.
    n_edges : int
        Number of edges of the graph.
    """

    _EDGE_DUPLICATION = 2
    _EDGE_SIZE = _UN_EDGE_SIZE
    _LL_EDGE_POS = _LL_UN_EDGE_POS

    def _add_edges(self, edges: np.ndarray) -> None:
        self._empty_edge_idx, self._n_edges = _add_undirected_edges(
            self._edges_buffer,
            edges,
            self._empty_edge_idx,
            self._n_edges,
            self._node2edges,
        )

    def get_edges(
        self, nodes: Optional[ArrayLike] = None, mode: str = 'indices'
    ) -> Union[List[np.ndarray], np.ndarray]:
        """Return the edges data of the given nodes.

        If no nodes are provided, all edges are returned.

        NOTE: when `nodes` is None, the returned edges are duplicated, so that
        if (u, v) was inserted this function will return both (u, v) and
        (v, u).

        Parameters
        ----------
        nodes : Optional[ArrayLike], optional
            Node indices, by default None
        mode : str
            Type of data queried from the edges. For example, `indices` or
            `coords`.

        Returns
        -------
        List[np.ndarray]
            List of (N_i) x 2 x D arrays, where N_i is the number of edges at
            the ith node.  D is the dimensionality of `coords` when mode ==
            `coords` and it's ignored when mode == `indices`. N_i dimension is
            ignored when N_i is 1.
        """
        return self._iterate_edges_generic(
            nodes,
            node2edges=self._node2edges,
            iterate_edges_func=_iterate_undirected_edges,
            mode=mode,
        )

    def _remove_edges(self, edges: np.ndarray) -> None:
        self._empty_edge_idx, self._n_edges = _remove_undirected_edges(
            edges,
            self._empty_edge_idx,
            self._n_edges,
            self._edges_buffer,
            self._node2edges,
        )

    def _remove_incident_edges(self, node_buffer_index: int) -> None:
        (
            self._empty_edge_idx,
            self._n_edges,
        ) = _remove_undirected_incident_edges(
            node_buffer_index,
            self._empty_edge_idx,
            self._n_edges,
            self._edges_buffer,
            self._node2edges,
        )



from napari_graph.base_graph import BaseGraph
from napari_graph.directed_graph import DirectedGraph
from napari_graph.interops import to_napari_graph
from napari_graph.undirected_graph import UndirectedGraph

try:
    from napari_graph._version import version as __version__
except ImportError:
    __version__ = "not-installed"

__all__ = [
    "BaseGraph",
    "DirectedGraph",
    "UndirectedGraph",
    "to_napari_graph",
]



from __future__ import annotations

import warnings
from abc import abstractmethod
from typing import Callable, List, Optional, Tuple, Union

import networkx as nx
import numpy as np
import pandas as pd
from numpy.typing import ArrayLike

from napari_graph.numba import njit, typed, types

"""
_NODE_EMPTY_PTR is used to fill the values of uninitialized/empty/removed nodes
"""
_NODE_EMPTY_PTR = -1

"""
_EDGE_EMPTY_PTR is used to fill the values of uninitialized/empty/removed edges
"""
_EDGE_EMPTY_PTR = -1


@njit(inline='always')
def _remove_edge(
    src_node: int,
    tgt_node: int,
    empty_idx: int,
    edges_buffer: np.ndarray,
    node2edges: np.ndarray,
    edge_size: int,
    ll_edge_pos: int,
) -> int:
    """Generic function to remove directed or undirected nodes.

    An additional removal of the target edges linked list is necessary for
    directed edges.

    Parameters
    ----------
    src_node : int
        Source node buffer index.
    tgt_node : int
        Target node buffer index.
    empty_idx : int
        First index of empty edges linked list.
    edges_buffer : np.ndarray
        Buffer of edges data.
    node2edges : np.ndarray
        Mapping from node indices to edge buffer indices -- head of edges linked list.
    edge_size : int
        Size of the edges on the buffer. It should be inlined when compiled.
    ll_edge_pos : int
        Position (shift) of the edge linked list on the edge buffer. It should
        be inlined when compiled.

    Returns
    -------
    int
        New first index of empty edges linked list.
    """

    idx = node2edges[src_node]
    prev_buffer_idx = _EDGE_EMPTY_PTR

    # safe guard against a corrupted buffer causing an infite loop
    for _ in range(edges_buffer.shape[0] // edge_size):
        if idx == _EDGE_EMPTY_PTR:
            raise ValueError("Could not find/remove edge.")

        buffer_idx = idx * edge_size
        next_edge_idx = edges_buffer[buffer_idx + ll_edge_pos]

        # edge found
        if edges_buffer[buffer_idx + 1] == tgt_node:
            # skipping found edge from linked list
            if prev_buffer_idx == _EDGE_EMPTY_PTR:
                node2edges[src_node] = next_edge_idx
            else:
                edges_buffer[prev_buffer_idx + ll_edge_pos] = next_edge_idx

            # clean up not necessary but good practice
            edges_buffer[buffer_idx : buffer_idx + edge_size] = _EDGE_EMPTY_PTR
            edges_buffer[buffer_idx + ll_edge_pos] = empty_idx

            break

        # moving to next edge
        idx = next_edge_idx
        prev_buffer_idx = buffer_idx
    else:
        raise ValueError(
            "Infinite loop detected at edge removal, edges buffer must be corrupted."
        )

    return idx


@njit(inline='always')
def _iterate_edges(
    edge_ptr_indices: np.ndarray,
    edges_buffer: np.ndarray,
    edge_size: int,
    ll_edge_pos: int,
) -> typed.List:
    """Iterate over the edges linked lists given their starting edges.

    It returns list of multiplicity 2, where each pair is an edge.

    Parameters
    ----------
    edge_ptr_indices : np.ndarray
        Array of starting indices.
    edges_buffer : np.ndarray
        Edges buffer.
    edge_size : int
        Size of the edges on the buffer. It should be inlined when compiled.
    ll_edge_pos : int
        Position (shift) of the edge linked list on the edge buffer. It should
        be inlined when compiled.

    Returns
    -------
    typed.List
        List of lists edges, adjacent nodes are at indices (k, k+1) such that k
        is even.
    """
    edges_list = typed.List()

    for idx in edge_ptr_indices:
        edges = typed.List.empty_list(types.int64)
        edges_list.append(edges)

        while idx != _EDGE_EMPTY_PTR:
            buffer_idx = idx * edge_size
            edges.append(edges_buffer[buffer_idx])  # src
            edges.append(edges_buffer[buffer_idx + 1])  # tgt
            idx = edges_buffer[buffer_idx + ll_edge_pos]

    return edges_list


@njit
def _contains_keys(
    map: typed.Dict,
    keys: np.ndarray,
) -> bool:
    """Returns true if at least one `key` is present on `map`."""
    for k in keys:
        if k in map:
            return True
    return False


@njit
def _update_world2buffer(
    world2buffer: typed.Dict,
    world_idx: np.ndarray,
    buffer_idx: np.ndarray,
) -> None:
    """Updates `world_idx` (keys) and `buffer_idx` (values) to `world2buffer` mapping."""
    for w, b in zip(world_idx, buffer_idx):
        world2buffer[w] = b


@njit
def _vmap_world2buffer(
    world2buffer: typed.Dict, world_idx: np.ndarray
) -> typed.Dict:
    """Maps world indices to buffer indices."""
    buffer_idx = np.empty(world_idx.shape[0], dtype=types.int64)
    for i in range(world_idx.shape[0]):
        buffer_idx[i] = world2buffer[world_idx[i]]
    return buffer_idx


class BaseGraph:
    """Abstract base graph class.

    Parameters
    ----------
    edges : ArrayLike
        Nx2 array of pair of nodes (edges).
    coords :
        Optional array of spatial coordinates of nodes.
    ndim : int
        Number of spatial dimensions of graph.
    n_nodes : int
        Optional number of nodes to pre-allocate in the graph.
    n_edges : int
        Optional number of edges to pre-allocate in the graph.
    """

    # abstract constants
    _EDGE_DUPLICATION: int
    _EDGE_SIZE: int
    _LL_EDGE_POS: int

    # allocation constants
    _ALLOC_MULTIPLIER = 1.1
    _ALLOC_MIN = 25

    def __init__(
        self,
        edges: ArrayLike = (),
        coords: Optional[Union[pd.DataFrame, ArrayLike]] = None,
        ndim: Optional[int] = None,
        n_nodes: Optional[int] = None,
        n_edges: Optional[int] = None,
    ):
        # validate nodes
        if coords is not None:
            if not isinstance(coords, pd.DataFrame):
                coords = pd.DataFrame(coords)
            if not np.issubdtype(coords.index.dtype, np.integer):
                raise ValueError(
                    f"The index of `coords` (data type: {coords.index.dtype}) must be an integer."
                )

            # validate nodes: ndim
            if len(coords.index) > 0:
                if ndim is not None:
                    if ndim != len(coords.columns):
                        raise ValueError(
                            f"`ndim` ({ndim}) does not match the number of columns in `coords` ({len(coords.columns)})."
                        )
                else:
                    ndim = len(coords.columns)

            # validate nodes: n_nodes
            if n_nodes is not None:
                if n_nodes < len(coords.index):
                    raise ValueError(
                        f"`n_nodes` ({n_nodes}) must be greater or equal than `coords` length ({len(coords.index)})."
                    )
            else:
                n_nodes = len(coords.index)

        # initialize nodes
        if n_nodes is None:
            n_nodes = self._ALLOC_MIN
        else:
            n_nodes = max(n_nodes, self._ALLOC_MIN)

        self._init_node_buffers(n_nodes)

        if ndim is not None:
            self._coords = np.empty((n_nodes, ndim), dtype=np.float32)
        else:
            self._coords = None

        if coords is not None:
            assert self._coords is not None
            self.add_nodes(indices=coords.index, coords=coords)

        # validate edges
        edges = np.asarray(edges)
        if len(edges) > 0:
            if edges.ndim != 2:
                raise ValueError(
                    f"`edges` ({edges.ndim} dimensions) must have 2 dimensions."
                )
            if edges.shape[1] != 2:
                raise ValueError(
                    f"`edges` (shape: {edges.shape}) must have shape E x 2."
                )

        # validate edges: n_edges
        if n_edges is not None:
            if n_edges < len(edges):
                raise ValueError(
                    f"`n_edges` ({n_edges}) must be greater or equal than `edges` length ({len(edges)}."
                )
        else:
            n_edges = len(edges)

        n_edges = max(n_edges, self._ALLOC_MIN)

        # initialize edges
        self._init_edge_buffers(n_edges)
        if len(edges) > 0:
            if coords is None:
                self.add_nodes(indices=np.unique(edges))
            self.add_edges(edges)

    def _init_node_buffers(self, n_nodes: int) -> None:
        self._empty_nodes: List[int] = list(reversed(range(n_nodes)))
        self._node2edges = np.full(
            n_nodes, fill_value=_EDGE_EMPTY_PTR, dtype=int
        )
        self._world2buffer = typed.Dict.empty(types.int64, types.int64)
        self._buffer2world = np.full(
            n_nodes, fill_value=_NODE_EMPTY_PTR, dtype=int
        )

    def _init_edge_buffers(self, n_edges: int) -> None:
        # if condition just to be safe, in case MIN_ALLOC is set to 0
        self._empty_edge_idx = 0 if n_edges > 0 else _EDGE_EMPTY_PTR
        self._n_edges = 0
        self._edges_buffer = np.full(
            n_edges * self._EDGE_DUPLICATION * self._EDGE_SIZE,
            fill_value=_EDGE_EMPTY_PTR,
            dtype=int,
        )
        self._edges_buffer[
            self._LL_EDGE_POS : -self._EDGE_SIZE : self._EDGE_SIZE
        ] = np.arange(1, self._EDGE_DUPLICATION * n_edges)

    @property
    def ndim(self) -> int:
        return self._coords.shape[1]

    @property
    def n_nodes(self) -> int:
        """Number of nodes in use."""
        return self.n_allocated_nodes - self.n_empty_nodes

    @property
    def n_allocated_nodes(self) -> int:
        """Number of total allocated nodes."""
        return len(self._buffer2world)

    @property
    def n_empty_nodes(self) -> int:
        """Number of nodes allocated but not used."""
        return len(self._empty_nodes)

    def get_nodes(self) -> np.ndarray:
        """Indices of graph nodes."""
        return self._buffer2world[self._buffer2world != _NODE_EMPTY_PTR]

    def get_coordinates(
        self, node_indices: Optional[ArrayLike] = None
    ) -> np.ndarray:
        """Coordinates of the given nodes.

        If none is provided it returns the coordinates of every node.
        """
        if self._coords is None:
            raise ValueError(
                "`get_coordinates` is only available for spatial graphs."
            )

        node_indices = self._validate_nodes(node_indices)
        node_indices = self._map_world2buffer(node_indices)
        return self._coords[node_indices]

    def _realloc_nodes_buffers(self, size: int) -> None:
        """Rellocs the nodes buffers and copies existing data.

        NOTE: Currently, only increasing the buffers' size is implemented.

        Parameters
        ----------
        size : int
            New buffer size.
        """
        prev_size = self.n_allocated_nodes
        size_diff = size - prev_size

        if size_diff < 0:
            raise NotImplementedError(
                "Node buffers size decrease not implemented."
            )

        elif size_diff == 0:
            raise ValueError("Tried to realloc to current buffer size.")

        if self._coords is not None:
            self._coords.resize(
                (size, self._coords.shape[1]), refcheck=False
            )  # zero-filled

        self._node2edges = np.append(
            self._node2edges,
            np.full(size_diff, fill_value=_EDGE_EMPTY_PTR, dtype=int),
        )
        self._buffer2world = np.append(
            self._buffer2world,
            np.full(size_diff, fill_value=_NODE_EMPTY_PTR, dtype=int),
        )
        self._empty_nodes = list(reversed(range(prev_size, size)))

    def get_next_valid_indices(self, count: int) -> ArrayLike:
        if count <= 0:
            raise ValueError(
                f"`count` must be a positive integer. Found {count}"
            )

        next_indices = self._buffer2world.max() + 1
        return np.arange(next_indices, next_indices + count)

    def add_nodes(
        self,
        *,
        indices: Optional[ArrayLike] = None,
        coords: Optional[ArrayLike] = None,
        count: Optional[int] = None,
    ) -> ArrayLike:
        """
        Add nodes to graph, at least one of the arguments must be supplied.
        `count` cannot be supplied with other arguments.

        Parameters
        ----------
        index : int
            Node index.
        coords : np.ndarray
            Node coordinates, optional for non-spatial graph.
        count : int:
            Number of nodes to be added.

        Returns
        -------
        ArrayLike
            Added nodes indices.
        """
        if count is not None and (indices is not None or coords is not None):
            raise ValueError(
                "`count` cannot be supplied with `indices` and `coords`."
            )

        if count is None and indices is None and coords is None:
            raise ValueError(
                "One of `indices`, `coords` or `count` must be supplied."
            )

        if coords is not None:
            coords = np.atleast_2d(coords)

        if indices is None:
            if count is None:
                count = coords.shape[0]
            indices = self.get_next_valid_indices(count)

        indices = np.atleast_1d(indices)
        if indices.ndim > 1:
            raise ValueError(
                f"`indices` must be 1-dimensional. Found {indices.ndim}."
            )

        if (self._coords is None) != (coords is None):
            if coords is None:
                raise ValueError(
                    "`coords` must be provided for spatial graphs."
                )
            else:
                raise ValueError(
                    "`coords` cannot be provided for non-spatial graphs."
                )

        if _contains_keys(self._world2buffer, indices):
            raise ValueError(
                f"One of the nodes {indices} are already present in the buffer."
            )

        if self.n_empty_nodes < len(indices):
            self._realloc_nodes_buffers(
                self._get_alloc_size(self.n_nodes + len(indices))
            )

        # flipping since _empty_nodes is a stack
        buffer_indices = np.flip(self._empty_nodes[-len(indices) :])

        if coords is not None:
            if indices.shape[0] != coords.shape[0]:
                raise ValueError(
                    f"`indices` and `coords` must be equal. Found {len(indices)} and {len(coords)}."
                )

            self._coords[buffer_indices] = coords

        _update_world2buffer(self._world2buffer, indices, buffer_indices)
        self._empty_nodes = self._empty_nodes[: -len(indices)]
        self._buffer2world[buffer_indices] = indices

        return indices

    def _get_alloc_size(self, size: int) -> int:
        return int(max(size * self._ALLOC_MULTIPLIER, self._ALLOC_MIN))

    def remove_node(self, index: int, is_buffer_domain: bool = False) -> None:
        """Remove node of given `index`, by default it's the world index.

        Parameters
        ----------
        index : int
            node index
        is_buffer_domain : bool, optional
            indicates if the index in on the buffer domain, by default False
        """
        if is_buffer_domain:
            index = self._buffer2world[index]
        buffer_index = self._world2buffer.pop(index)
        self._remove_incident_edges(buffer_index)
        self._buffer2world[buffer_index] = _NODE_EMPTY_PTR
        self._empty_nodes.append(buffer_index)

    @abstractmethod
    def _remove_incident_edges(self, node_buffer_index: int) -> None:
        """Abstract method, removes node at given buffer index."""
        raise NotImplementedError

    def _realloc_edges_buffers(self, n_edges: int) -> None:
        """Reallocs the edges buffer and copies existing data.

        NOTE: Currently, only increasing the buffers' size is implemented.

        Parameters
        ----------
        n_edges : int
            New number of edges.
        """

        # augmenting size to match dummy edges
        size = n_edges * self._EDGE_DUPLICATION
        prev_size = self.n_allocated_edges * self._EDGE_DUPLICATION
        diff_size = size - prev_size

        if diff_size < 0:
            raise NotImplementedError(
                "Edge buffer size decrease not implemented."
            )
        elif diff_size == 0:
            raise ValueError(
                f"Tried to realloc to current buffer size ({self.n_allocated_edges})."
            )

        prev_buffer_size = len(self._edges_buffer)

        self._edges_buffer = np.append(
            self._edges_buffer,
            np.full(
                diff_size * self._EDGE_SIZE,
                fill_value=_EDGE_EMPTY_PTR,
                dtype=int,
            ),
        )

        # fills empty edges ptr
        self._edges_buffer[
            prev_buffer_size
            + self._LL_EDGE_POS : -self._EDGE_SIZE : self._EDGE_SIZE
        ] = np.arange(prev_size + 1, size)

        # appends existing empty edges linked list to the end of the new list
        self._edges_buffer[
            self._LL_EDGE_POS - self._EDGE_SIZE
        ] = self._empty_edge_idx
        self._empty_edge_idx = prev_size

    @property
    def n_allocated_edges(self) -> int:
        """Number of total allocated edges."""
        return len(self._edges_buffer) // (
            self._EDGE_DUPLICATION * self._EDGE_SIZE
        )

    @property
    def n_empty_edges(self) -> int:
        """Number of allocated edges but not used."""
        return self.n_allocated_edges - self.n_edges

    @property
    def n_edges(self) -> int:
        """Number of edges in use."""
        return self._n_edges

    def _validate_nodes(
        self, node_indices: Optional[ArrayLike] = None
    ) -> np.ndarray:
        """Converts and validates the nodes indices."""

        # NOTE: maybe the nodes could be mappend inside this function
        if node_indices is None:
            return self.get_nodes()

        node_indices = np.atleast_1d(node_indices)

        if not np.issubdtype(node_indices.dtype, np.integer):
            raise ValueError(
                f"Node indices must be integer. Found {node_indices.dtype}."
            )

        if node_indices.ndim != 1:
            raise ValueError(
                "Node indices must be 1-dimensional. "
                f"Found {node_indices.ndim}-dimensional."
            )

        return node_indices

    def _validate_edges(self, edges: ArrayLike) -> np.ndarray:
        """Converts and validates the edges."""
        edges = np.atleast_2d(edges)

        if not np.issubdtype(edges.dtype, np.integer):
            raise ValueError(f"Edges must be integer. Found {edges.dtype}.")

        if edges.ndim != 2:
            raise ValueError(
                "Edges must be 1- or 2-dimensional. "
                f"Found {edges.ndim}-dimensional."
            )

        if edges.shape[1] != 2:
            raise ValueError(
                f"Edges must be a sequence of length 2 arrays. "
                f"Found length {edges.shape[1]}"
            )

        return edges

    @abstractmethod
    def _add_edges(self, edges: np.ndarray) -> None:
        """Abstract method.

        Requires different implementation for undirected and directed graphs.
        """
        raise NotImplementedError

    def add_edges(self, edges: ArrayLike) -> None:
        """Add edges into the graph.

        TODO : add parameter for edges features

        Parameters
        ----------
        edges : ArrayLike
            A list of 2-dimensional tuples or an Nx2 array with a pair of
            node indices.
        """
        edges = self._validate_edges(edges)

        if self.n_empty_edges < len(edges):
            self._realloc_edges_buffers(
                self._get_alloc_size(self.n_edges + len(edges))
            )

        self._add_edges(self._map_world2buffer(edges))

    @abstractmethod
    def _remove_edges(self, edges: np.ndarray) -> None:
        raise NotImplementedError

    def remove_edges(self, edges: ArrayLike) -> None:
        """Remove edges from the graph.

        Parameters
        ----------
        edges : ArrayLike
            A list of 2-dimensional tuples or an Nx2 array with a pair of
            node indices.
        """
        edges = self._validate_edges(edges)
        edges = self._map_world2buffer(edges)
        self._remove_edges(edges)

    def _map_world2buffer(self, world_idx: np.ndarray) -> np.ndarray:
        """Flatten the world indices buffer maps into buffer coordinates.

        ... and reshape back to original space.
        """
        shape = world_idx.shape
        buffer_idx = _vmap_world2buffer(
            self._world2buffer, world_idx.reshape(-1)
        )
        return buffer_idx.reshape(shape)

    def _iterate_edges(
        self,
        node_world_indices: ArrayLike,
        node2edges: np.ndarray,
        iterate_edges_func: Callable[
            [np.ndarray, np.ndarray], List[np.ndarray]
        ],
    ) -> List[List]:
        """Helper function to iterate over edges and return buffer indices.

        Parameters
        ----------
        node_world_indices : ArrayLike
            Nodes world indices.
        node2edges : np.ndarray
            Mapping from node indices to edge buffer indices -- head of edges linked list.
        iterate_edges_func : [np.ndarray, np.ndarray] -> List[np.ndarray]
            Function that iterates the edges from `edges_ptr_indices` and
            `edges_buffer`.

        Returns
        -------
        List[List]
            List of Lists of length 2 * N_i, where N_i is the number of edges
            at the ith node.
        """
        node_world_indices = self._validate_nodes(node_world_indices)

        flat_edges = iterate_edges_func(
            node2edges[self._map_world2buffer(node_world_indices)],
            self._edges_buffer,
        )
        return flat_edges

    def _iterate_edges_generic(
        self,
        node_world_indices: ArrayLike,
        node2edges: np.ndarray,
        iterate_edges_func: Callable[
            [np.ndarray, np.ndarray], List[np.ndarray]
        ],
        mode: str,
    ) -> Union[List[np.ndarray], np.ndarray]:
        """Iterate over any kind of edges and return their world indices.

        Parameters
        ----------
        node_world_indices : ArrayLike
            Nodes world indices.
        node2edges : np.ndarray
            Mapping from node indices to edge buffer indices -- head of edges linked list.
        iterate_edges_func : [np.ndarray, np.ndarray] -> List[np.ndarray]
            Function that iterates the edges from `edges_ptr_indices` and
            `edges_buffer`.
        mode : str
            Type of data queried from the edges. For example, `indices` or
            `coords`.

        Returns
        -------
        List[np.ndarray]
            List of N_i x 2 x D arrays, where N_i is the number of edges at
            the ith node.  D is the dimensionality of `coords` when
            mode == `coords` and it's ignored when mode == `indices`.
        """
        if mode.lower() == 'coords' and self._coords is None:
            raise ValueError(
                "`coords` mode only available for spatial graphs."
            )

        flat_edges = self._iterate_edges(
            node_world_indices, node2edges, iterate_edges_func
        )

        if mode.lower() == 'indices':
            edges_data = [
                self._buffer2world[e].reshape(-1, 2)
                if len(e) > 0
                else np.empty((0, 2))
                for e in flat_edges
            ]
        elif mode.lower() == 'coords':
            ndim = self._coords.shape[1]
            edges_data = [
                self._coords[e].reshape(-1, 2, ndim)
                if len(e) > 0
                else np.empty((0, 2, ndim))
                for e in flat_edges
            ]
        # NOTE: here `mode` could also query the edges features.
        # Not implemented yet.
        else:
            modes = ('indices', 'coords')
            raise ValueError(
                f"Edge iteration mode not found. Received {mode}, "
                f"expected {modes}."
            )

        if len(edges_data) == 1:
            return edges_data[0]
        else:
            return edges_data

    @abstractmethod
    def get_edges(
        self, nodes: Optional[ArrayLike] = None, mode: str = 'indices'
    ) -> Union[List[np.ndarray], np.ndarray]:
        raise NotImplementedError

    def get_edges_buffers(
        self, is_buffer_domain: bool = False
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Return valid edges in buffer or world domain.

        Return the indices (buffer domain) and the (source, target) values
        (world domain) of all valid edges.

        Undirected edges are not duplicated.

        This function is useful for loading the data for visualization.

        Parameters
        ----------
        is_buffer_domain : bool
            flag indicating if it should return `world` or `buffer` domain.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            Buffer indices (buffer domain) and (source, target) (world domain by default).
        """
        unique_edge_size = self._EDGE_SIZE * self._EDGE_DUPLICATION
        buffer_size = len(self._edges_buffer)
        indices = np.arange(0, buffer_size, unique_edge_size)

        # reshaping such that each row is (source id, target id, ...)
        buffer = self._edges_buffer.reshape((-1, unique_edge_size))
        edges = buffer[:, :2]  # (source, target)

        valid = edges[:, 0] != _EDGE_EMPTY_PTR

        indices = indices[valid]
        edges = edges[valid]
        if not is_buffer_domain:
            edges = self._buffer2world[edges]

        return indices, edges

    def __len__(self) -> int:
        """Number of nodes in use."""
        return self.n_nodes

    def initialized_buffer_mask(self) -> np.ndarray:
        """Compute mask of nodes that have already been initialized.

        Returns
        -------
        np.ndarray
            Boolean array of valid node, it has the same length the buffer.
        """
        return self._buffer2world != _NODE_EMPTY_PTR

    @property
    def coords_buffer(self) -> np.ndarray:
        """Returns the actual coordinates buffer. It's not a copy."""
        if self._coords is None:
            raise ValueError(
                "graph does not have a `coords` attribute. "
                "It is not a spatial graph."
            )
        return self._coords

    def is_spatial(self) -> bool:
        """True if self is a spatial graph (has coordinates attribute)."""
        return self._coords is not None

    @staticmethod
    def from_networkx(graph: nx.Graph) -> BaseGraph:
        """Loads a Directed or Undirected napari-graph from a NetworkX graph.

        Parameters
        ----------
        graph : nx.Graph
            The NetworkX graph to be converted.
        """
        from napari_graph.directed_graph import DirectedGraph
        from napari_graph.undirected_graph import UndirectedGraph

        nodes = np.array(list(graph.nodes()))
        if not np.issubdtype(nodes.dtype, np.integer) or nodes.ndim > 1:
            graph_int_nodes = nx.convert_node_labels_to_integers(
                graph, label_attribute='_node_id'
            )
            warnings.warn(
                'Node IDs must be integers. They have been converted '
                'automatically.'
            )
        else:
            graph_int_nodes = graph
        coords_dict = nx.get_node_attributes(graph_int_nodes, "pos")
        if len(coords_dict) > 0:
            coords_df = pd.DataFrame.from_dict(coords_dict, orient="index")
        else:
            coords_df = None

        edges = graph_int_nodes.edges
        if len(edges) > 0:
            edges = np.atleast_2d(edges)

        return (
            DirectedGraph(edges, coords_df)
            if graph_int_nodes.is_directed()
            else UndirectedGraph(edges, coords_df)
        )

    def to_networkx(self) -> nx.Graph:
        """Convert it self into NetworkX graph.

        Parameters
        ----------
        graph : BaseGraph
            napari-graph Graph

        Returns
        -------
        nx.Graph
            An equivalent NetworkX graph.
        """
        from napari_graph.directed_graph import DirectedGraph

        if isinstance(self, DirectedGraph):
            out_graph = nx.DiGraph()
        else:
            out_graph = nx.Graph()

        if self.is_spatial():
            for node_id, pos in zip(self.get_nodes(), self.get_coordinates()):
                # note: some nx functions are unhappy with arrays in node
                # attributes because you can't compare arrays with ==.
                # So one day we might want to cast to tuple.
                out_graph.add_node(node_id, pos=pos)
        else:
            out_graph.add_nodes_from(self.get_nodes())

        edges = self.get_edges()
        if isinstance(edges, list) and len(edges) > 0:
            edges = np.concatenate(edges, axis=0)

        edges_as_tuples = list(map(tuple, edges))
        out_graph.add_edges_from(edges_as_tuples)

        return out_graph

    def subgraph_edges(
        self,
        node_indices: ArrayLike,
        is_buffer_domain: bool = False,
    ) -> ArrayLike:
        """Returns edges (node pair) where both nodes are presents.

        Parameters
        ----------
        nodes_indices : np.ndarray
            Subset of nodes used for selection.
        is_buffer_domain : bool
            When true `node_indices` and returned edges are on buffer domain.

        Returns
        -------
        np.ndarray
            (N x 2) array of nodes indices, where N is the number of valid edges from the induced subgraph.
        """
        _, edges = self.get_edges_buffers(is_buffer_domain)

        if is_buffer_domain:
            mask = np.zeros(self._buffer2world.shape[0], dtype=bool)
            mask[node_indices] = True
            subgraph_edges = edges[mask[edges[:, 0]] & mask[edges[:, 1]]]

        else:
            mask = np.isin(edges, node_indices).all(axis=1)
            assert mask.shape[0] == edges.shape[0]
            subgraph_edges = edges[mask]

        return subgraph_edges



from itertools import product
from typing import Any, Callable, List, Optional, Protocol, Type

import numpy as np
import pandas as pd
import pytest
from numpy.typing import ArrayLike

from napari_graph import DirectedGraph, UndirectedGraph
from napari_graph.base_graph import _EDGE_EMPTY_PTR, BaseGraph
from napari_graph.undirected_graph import _LL_UN_EDGE_POS, _UN_EDGE_SIZE


@pytest.mark.parametrize("n_prealloc_edges", [0, 2, 5])
def test_undirected_edge_addition(n_prealloc_edges: int) -> None:
    coords = pd.DataFrame(
        [
            [0, 2.5],
            [4, 2.5],
            [1, 0],
            [2, 3.5],
            [3, 0],
        ],
        columns=["y", "x"],
    )

    edges = [[0, 1], [1, 2], [2, 3], [3, 4], [0, 4]]

    graph = UndirectedGraph(
        coords=coords,
        n_edges=n_prealloc_edges,
    )
    graph.add_edges(edges)

    for node_idx, node_edges in zip(
        coords.index, graph.get_edges(coords.index)
    ):
        # checking if two edges per node and connecting only two nodes
        assert node_edges.shape == (2, 2)

        # checking if the given index is the source
        assert np.all(node_edges[:, 0] == node_idx)

        # checking if the edges are corrected
        for edge in node_edges:
            assert sorted(edge) in edges


@pytest.mark.parametrize("n_prealloc_edges", [0, 2, 5])
def test_directed_edge_addition(n_prealloc_edges: int) -> None:
    coords = pd.DataFrame(
        [
            [0, 2.5],
            [4, 2.5],
            [1, 0],
            [2, 3.5],
            [3, 0],
        ],
        columns=["y", "x"],
    )

    edges = np.asarray([[0, 1], [1, 2], [2, 3], [3, 4], [4, 0]])

    graph = DirectedGraph(
        coords=coords,
        n_edges=n_prealloc_edges,
    )
    graph.add_edges(edges)

    source_edges = np.asarray(graph.get_source_edges(coords.index))
    target_edges = np.asarray(
        graph.get_target_edges(np.roll(coords.index, -1))
    )
    assert np.all(source_edges == edges[:, np.newaxis, :])
    assert np.all(target_edges == edges[:, np.newaxis, :])


@pytest.mark.parametrize("n_prealloc_nodes", [0, 3, 6, 12])
def test_node_addition_indices_coords(n_prealloc_nodes: int) -> None:
    # test node addition with indices and coords and different pre-allocation size
    size = 6
    ndim = 3

    indices = np.random.choice(range(100), size=size, replace=False)
    coords = np.random.randn(size, ndim)

    graph = DirectedGraph(ndim=ndim, n_nodes=n_prealloc_nodes)
    for i in range(size):
        graph.add_nodes(indices=indices[i], coords=coords[i])
        assert len(graph) == i + 1

    np.testing.assert_allclose(graph._coords[: len(graph)], coords)
    np.testing.assert_array_equal(graph._buffer2world[: len(graph)], indices)
    np.testing.assert_array_equal(
        graph._map_world2buffer(indices), range(size)
    )


def test_node_addition_non_spatial() -> None:
    graph = DirectedGraph()

    size = 5
    new_indices = graph.add_nodes(count=size)

    np.testing.assert_array_equal(new_indices, np.arange(size))

    new_indices = graph.add_nodes(indices=[10, 11])
    nodes = graph.get_nodes()

    np.testing.assert_equal(new_indices, [10, 11])
    np.testing.assert_array_equal(nodes[:size], np.arange(size))
    np.testing.assert_array_equal(nodes[size:], [10, 11])

    with pytest.raises(ValueError):
        graph.add_nodes(coords=[[0, 1], [2, 3]])

    graph.add_nodes(indices=15)
    assert len(graph) == 8

    with pytest.raises(ValueError):
        graph.add_nodes(indices=15)


def test_node_addition_spatial() -> None:
    graph = DirectedGraph(ndim=2)

    with pytest.raises(ValueError):
        graph.add_nodes(indices=5, coords=[0, 0], count=1)

    new_index = graph.add_nodes(coords=[2, 2])
    assert len(graph) == 1
    assert new_index == 0
    np.testing.assert_equal(graph.get_coordinates(), [[2, 2]])

    new_index = graph.add_nodes(indices=[10, 11], coords=[[0, 0], [1, 1]])
    assert len(graph) == 3
    np.testing.assert_equal(new_index, [10, 11])

    with pytest.raises(ValueError):
        graph.add_nodes(indices=13)

    with pytest.raises(ValueError):
        graph.add_nodes(count=2)

    with pytest.raises(ValueError):
        graph.add_nodes(indices=[2, 3], coords=[[1, 2]])

    assert len(graph) == 3


@pytest.mark.parametrize(
    "graph_type,ndim",
    product([UndirectedGraph, DirectedGraph], [None, 2, 3]),
)
def test_empty_graph(graph_type: Type[BaseGraph], ndim: Optional[int]) -> None:
    _ = graph_type(ndim=ndim)


class TestGraph:
    _GRAPH_CLASS: Type[BaseGraph]
    __test__ = False  # ignored for testing
    _index_shift = 0  # shift used to test special indexing

    def setup_method(self, method: Callable) -> None:
        self.coords = pd.DataFrame(
            [
                [0, 2.5],
                [4, 2.5],
                [1, 0],
                [2, 3.5],
                [3, 0],
            ],
            index=np.arange(5) + self._index_shift,
            columns=["y", "x"],
        )

        self.edges = (
            np.asarray([[0, 1], [1, 2], [2, 3], [3, 4], [4, 0]], dtype=int)
            + self._index_shift
        )

        self.graph = self._GRAPH_CLASS(
            edges=self.edges,
            coords=self.coords,
        )

    @staticmethod
    def contains(edge: ArrayLike, edges: List[ArrayLike]) -> bool:
        return any(
            np.allclose(e, edge) if len(e) > 0 else False for e in edges
        )

    def teardown_method(self, method: Callable) -> None:
        del self.edges, self.graph
        if hasattr(self, "coords"):
            del self.coords

    def test_edge_buffers(self) -> None:
        # testing buffer correctness on a non-trivial case when a node was removed
        node_id = 3 + self._index_shift
        self.graph.remove_node(node_id)

        valid_edges = np.logical_not(np.any(self.edges == node_id, axis=1))

        expected_edges = self.edges[valid_edges, :]
        (expected_indices,) = np.nonzero(valid_edges)
        expected_indices *= (
            self.graph._EDGE_SIZE * self.graph._EDGE_DUPLICATION
        )

        indices, edges = self.graph.get_edges_buffers()

        assert np.allclose(expected_indices, indices)
        assert np.allclose(expected_edges, edges)

    def test_subgraph_edges(self) -> None:
        # 3 is disconnected in subgraph
        nodes_ids = np.asarray([0, 1, 3]) + self._index_shift

        subgraph = self.graph.subgraph_edges(nodes_ids)
        expected_subgraph = np.asarray([[0, 1]]) + self._index_shift

        np.testing.assert_array_equal(subgraph, expected_subgraph)

        buffer_node_ids = np.asarray([0, 1, 3])
        buffer_subgraph = self.graph.subgraph_edges(
            buffer_node_ids, is_buffer_domain=True
        )

        np.testing.assert_equal(buffer_subgraph, [[0, 1]])


class TestDirectedGraph(TestGraph):
    graph: DirectedGraph  # required by typing
    _GRAPH_CLASS = DirectedGraph
    __test__ = True

    def test_edge_removal(self) -> None:
        edge = np.asarray([0, 1]) + self._index_shift
        self.graph.remove_edges(edge)
        assert self.graph.n_edges == 4
        assert self.graph.n_empty_edges == self.graph._ALLOC_MIN - 4
        assert not self.contains(edge, self.graph.get_source_edges())
        assert not self.contains(np.flip(edge), self.graph.get_target_edges())

        edges = np.asarray([[1, 2], [2, 3]]) + self._index_shift
        self.graph.remove_edges(edges)
        assert self.graph.n_edges == 2
        assert self.graph.n_empty_edges == self.graph._ALLOC_MIN - 2
        assert not self.contains(edges[0], self.graph.get_source_edges())
        assert not self.contains(edges[1], self.graph.get_source_edges())

        assert self.graph.n_allocated_edges == self.graph._ALLOC_MIN

    def test_node_removal(self) -> None:
        nodes = np.asarray([3, 4, 1]) + self._index_shift
        original_size = len(self.graph)

        for i in range(len(nodes)):
            node = nodes[i]
            self.graph.remove_node(node)

            for edge in self.graph.get_source_edges():
                assert node not in edge

            for edge in self.graph.get_target_edges():
                assert node not in edge

            assert node not in self.graph.get_nodes()
            assert len(self.graph) == original_size - i - 1

    def test_edge_coordinates(self) -> None:
        coords = self.coords.loc[self.edges.ravel()].to_numpy()
        coords = coords.reshape(self.edges.shape + (-1,))

        source_edge_coords = np.concatenate(
            self.graph.get_source_edges(mode='coords'), axis=0
        )
        assert np.allclose(coords, source_edge_coords)

        target_edges_coords = np.concatenate(
            self.graph.get_target_edges(mode='coords'), axis=0
        )
        rolled_coords = np.roll(coords, shift=1, axis=0)
        assert np.allclose(rolled_coords, target_edges_coords)


class TestUndirectedGraph(TestGraph):
    _GRAPH_CLASS = UndirectedGraph
    __test__ = True

    def test_edge_removal(self) -> None:
        edge = np.asarray([0, 1]) + self._index_shift
        self.graph.remove_edges(edge)
        assert self.graph.n_edges == 4
        assert self.graph.n_empty_edges == self.graph._ALLOC_MIN - 4
        assert not self.contains(edge, self.graph.get_edges())
        assert not self.contains(np.flip(edge), self.graph.get_edges())

        edges = np.asarray([[1, 2], [2, 3]]) + self._index_shift
        self.graph.remove_edges(edges)
        assert self.graph.n_edges == 2
        assert self.graph.n_empty_edges == self.graph._ALLOC_MIN - 2
        assert not self.contains(edges[0], self.graph.get_edges())
        assert not self.contains(edges[1], self.graph.get_edges())

        assert self.graph.n_allocated_edges == self.graph._ALLOC_MIN

        self.assert_empty_linked_list_pairs_are_neighbors()

    def test_node_removal(self) -> None:
        nodes = np.asarray([3, 4, 1]) + self._index_shift
        original_size = len(self.graph)

        for i in range(len(nodes)):
            node = nodes[i]
            self.graph.remove_node(node)

            for edge in self.graph.get_edges():
                assert node not in edge

            assert node not in self.graph.get_nodes()
            assert len(self.graph) == original_size - i - 1

        self.assert_empty_linked_list_pairs_are_neighbors()

    def test_edge_coordinates(self) -> None:
        edge_coords = self.graph.get_edges(mode='coords')

        for node, coords in zip(self.graph.get_nodes(), edge_coords):
            for i, edge in enumerate(self.graph.get_edges(node)):
                assert np.allclose(
                    self.coords.loc[edge, ["y", "x"]].to_numpy(), coords[i]
                )

    def assert_empty_linked_list_pairs_are_neighbors(self) -> None:
        # testing if empty edges linked list pairs are neighbors
        empty_idx = self.graph._empty_edge_idx
        while empty_idx != _EDGE_EMPTY_PTR:
            next_empty_idx = self.graph._edges_buffer[
                empty_idx * _UN_EDGE_SIZE + _LL_UN_EDGE_POS
            ]
            assert empty_idx + 1 == next_empty_idx
            # skipping one
            empty_idx = self.graph._edges_buffer[
                next_empty_idx * _UN_EDGE_SIZE + _LL_UN_EDGE_POS
            ]


class NonSpatialMixin(Protocol):
    # required by typing
    _GRAPH_CLASS: Any
    graph: BaseGraph
    edges: np.ndarray
    _index_shift: int

    def setup_method(self, method: Callable) -> None:
        self.edges = (
            np.asarray([[0, 1], [1, 2], [2, 3], [3, 4], [4, 0]], dtype=int)
            + self._index_shift
        )

        self.graph = self._GRAPH_CLASS(edges=self.edges)

    def test_edge_coordinates(self) -> None:
        pytest.skip("Non-spatial graph has no coordinates.")


class TestNonSpatialDirectedGraph(NonSpatialMixin, TestDirectedGraph):
    pass


class TestNonSpatialUnirectedGraph(NonSpatialMixin, TestUndirectedGraph):
    pass


class TestDirectedGraphSpecialIndex(TestDirectedGraph):
    _index_shift = 10


class TestUndirectedGraphSpecialIndex(TestUndirectedGraph):
    _index_shift = 10


@pytest.mark.parametrize(
    "node_id",
    [0, 1],
)
def test_removing_last_edge_from_digraph(node_id: int) -> None:
    # regression test from bug reported on Zulip:
    # https://napari.zulipchat.com/#narrow/stream/360030-working-group-graph/topic/Node.20removal.20bug/near/411290388
    coords = np.asarray([[0, 0], [20, 20], [100, 100], [120, 120]])

    graph = DirectedGraph(edges=[[0, 1]], coords=coords)

    # removing node 3 is fine since it is not connected to the sole edge
    graph.remove_node(3)

    # removing either 0 or 1 (source or destination) caused error
    graph.remove_node(node_id)

    assert graph.n_edges == 0
    assert graph.n_nodes == 2



import itertools
from typing import Any, Callable, List

import networkx as nx
import numpy as np
import pandas as pd
import pytest

from napari_graph import (  # noqa
    BaseGraph,
    DirectedGraph,
    UndirectedGraph,
    to_napari_graph,
)


def _graph_list() -> List[BaseGraph]:

    coords = pd.DataFrame(
        [
            [0, 2.5],
            [4, 2.5],
            [1, 0],
            [2, 3.5],
            [3, 0],
        ],
        columns=["y", "x"],
    )

    edges = np.asarray([[0, 1], [1, 2], [2, 3], [3, 4], [4, 0]])

    empty_graph = UndirectedGraph()

    graph = UndirectedGraph(edges)

    digraph = DirectedGraph(edges)

    spatial_graph = UndirectedGraph(
        edges=edges,
        coords=coords,
    )

    spatial_digraph = DirectedGraph(
        edges=edges,
        coords=coords,
    )

    only_coords = UndirectedGraph(coords=coords)

    only_coords_di = DirectedGraph(coords=coords)

    return [
        empty_graph,
        graph,
        digraph,
        spatial_graph,
        spatial_digraph,
        only_coords,
        only_coords_di,
    ]


@pytest.mark.parametrize(
    "in_graph,to_class",
    itertools.product(_graph_list(), [BaseGraph.to_networkx]),
)
def test_conversion(
    in_graph: BaseGraph, to_class: Callable[[BaseGraph], Any]
) -> None:

    nxgraph = to_class(in_graph)
    out_graph = to_napari_graph(nxgraph)

    assert np.array_equal(in_graph.get_nodes(), out_graph.get_nodes())

    if in_graph.is_spatial():
        assert np.array_equal(
            in_graph.get_coordinates(), out_graph.get_coordinates()
        )

    if in_graph.n_edges == 0:
        assert out_graph.n_edges == 0

    else:
        in_graph_edges = np.concatenate(in_graph.get_edges(), axis=0).sort()
        out_graph_edges = np.concatenate(out_graph.get_edges(), axis=0).sort()
        assert np.array_equal(in_graph_edges, out_graph_edges)


def test_weighted_networkx_graph() -> None:

    nxgraph = nx.DiGraph()
    nxgraph.add_edge(10, 11, weight=0.1)
    nxgraph.add_edge(11, 12, weight=0.2)
    nxgraph.add_edge(12, 10, weight=0.3)

    nxgraph_edges = np.asarray(nxgraph.edges).sort()

    graph = to_napari_graph(nxgraph)
    graph_edges = np.concatenate(graph.get_edges(), axis=0).sort()

    assert np.array_equal(nxgraph_edges, graph_edges)


def test_table_like_graphs() -> None:

    coords = pd.DataFrame(
        [
            [0, 2.5],
            [4, 2.5],
            [1, 0],
            [2, 3.5],
            [3, 0],
        ],
        columns=["y", "x"],
    )
    coords.index = np.arange(
        10, 5, -1
    )  # testing index that don't start with zero

    # testing pandas dataframe
    graph = to_napari_graph(coords)
    assert np.allclose(graph.get_coordinates(), coords)

    # testing numpy array
    graph = to_napari_graph(coords.to_numpy())
    assert np.allclose(graph.get_coordinates(), coords.to_numpy())

    # testing bad table
    not_a_table = np.ones((5, 5, 5))
    with pytest.raises(ValueError):
        graph = to_napari_graph(not_a_table)


def test_networkx_non_integer_ids():
    """Check that passing nx graph with non-integer IDs doesn't crash."""
    g = nx.hexagonal_lattice_graph(5, 5, with_positions=True)
    with pytest.warns(UserWarning, match='Node IDs must be integers.'):
        BaseGraph.from_networkx(g)


def test_networkx_basic_roundtrip():
    g = nx.hexagonal_lattice_graph(5, 5, with_positions=True)
    gint = nx.convert_node_labels_to_integers(g)
    ng = BaseGraph.from_networkx(gint)
    g2 = ng.to_networkx()
    # convert positions to tuples because nx comparison tools don't like arrays
    for node in g2.nodes():
        g2.nodes[node]['pos'] = tuple(g2.nodes[node]['pos'])
    assert nx.utils.edges_equal(gint.edges, g2.edges)
    assert set(gint.nodes) == set(g2.nodes)
    for node in gint.nodes:
        np.testing.assert_allclose(
            gint.nodes[node]['pos'], g2.nodes[node]['pos']
        )



from typing import Any

import networkx as nx
import numpy as np
import pandas as pd

from napari_graph.base_graph import BaseGraph
from napari_graph.undirected_graph import UndirectedGraph

COMPATIBLE_CLASSES = {
    BaseGraph: lambda x: x,
    nx.Graph: BaseGraph.from_networkx,
    pd.DataFrame: lambda x: UndirectedGraph(coords=x),
    np.ndarray: lambda x: UndirectedGraph(coords=x),
}


def to_napari_graph(graph: Any) -> BaseGraph:
    """Generic function to convert "any" graph to a napari-graph.

    Supported formats:
        - napari-graph (itself)
        - NetworkX

    Parameters
    ----------
    graph : Any
        Any kind of graph to a napari-graph. See supported formats above.

    Returns
    -------
    BaseGraph
        A napari-graph graph.
    """
    for cls, conversion_func in COMPATIBLE_CLASSES.items():
        if isinstance(graph, cls):
            return conversion_func(graph)

    raise NotImplementedError(
        f"Conversion from {type(graph)} to napari-graph does not exist."
    )
